"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.decodeMap = exports.encodeMap = exports.toJSON = exports.fromJSON = void 0;
/* eslint-disable @typescript-eslint/no-use-before-define */
const assert_1 = __importDefault(require("assert"));
const lodash_1 = __importDefault(require("lodash"));
const ethers_1 = require("ethers");
const buffer_1 = require("buffer");
const types_1 = require("./types");
function decodeValue(fieldType, fieldJsonValue) {
    if (fieldJsonValue === undefined) {
        return fieldJsonValue;
    }
    switch (fieldType.type) {
        case 'class': {
            return fieldType.value.fromJSON(types_1.JSONbigNative.stringify(fieldJsonValue));
        }
        case 'string': {
            return String(fieldJsonValue);
        }
        case 'address': {
            return ethers_1.ethers.utils.getAddress(fieldJsonValue);
        }
        case 'number': {
            return Number(fieldJsonValue);
        }
        case 'boolean': {
            return String(fieldJsonValue) === 'true';
        }
        case 'bigint': {
            return fieldJsonValue === null ? undefined : BigInt(fieldJsonValue);
        }
        case 'uint':
        case 'uint64': {
            return BigInt(fieldJsonValue);
        }
        case 'buffer': {
            return fieldJsonValue === null ? null : buffer_1.Buffer.from(fieldJsonValue, 'base64');
        }
        case 'object': {
            (0, assert_1.default)(fieldType.value);
            const objectTypeEncodingMap = fieldType.value;
            const objFieldValue = {};
            Object.keys(fieldJsonValue).forEach((key) => {
                const lowercaseFieldKey = lowercaseFirstLetter(key);
                objFieldValue[lowercaseFieldKey] = decodeValue(objectTypeEncodingMap[lowercaseFieldKey], fieldJsonValue[key]);
            });
            return objFieldValue;
        }
        case 'map': {
            (0, assert_1.default)(fieldType.key);
            (0, assert_1.default)(fieldType.value);
            return decodeMap(fieldType.key, fieldType.value, fieldJsonValue);
        }
        case 'array': {
            if (fieldJsonValue === null) {
                return null;
            }
            (0, assert_1.default)(fieldType.value);
            return fieldJsonValue.map((value) => decodeValue(fieldType.value, value));
        }
        default:
            throw new Error(`Unknown field type ${fieldType.type}`);
    }
}
// Go compatible JSON unmarshalling utility method
function fromJSON(jsonEncodingMap, data, keysMap = new Map()) {
    // Parse the JSON data string
    const jsonValue = types_1.JSONbigNative.parse(data);
    const props = {};
    Object.keys(jsonEncodingMap).forEach((fieldKey) => {
        var _a;
        const fieldType = jsonEncodingMap[fieldKey];
        // Use mapped key in props
        const propsKey = (_a = keysMap.get(fieldKey)) !== null && _a !== void 0 ? _a : fieldKey;
        const capitalizedFieldKey = capitalizeFirstLetter(fieldKey);
        props[propsKey] = decodeValue(fieldType, jsonValue[capitalizedFieldKey]);
    });
    return props;
}
exports.fromJSON = fromJSON;
// Go compatible JSON marshalling utility method
function toJSON(jsonEncodingMap, obj, keysMap = new Map()) {
    let mappedObj = Object.assign({}, obj);
    // Replace object keys with mapped & capitalized keys
    mappedObj = lodash_1.default.mapKeys(mappedObj, (value, key) => { var _a; return capitalizeFirstLetter((_a = keysMap.get(key)) !== null && _a !== void 0 ? _a : key); });
    // Create a new object having keys in order of jsonEncodingMap keys
    const jsonObj = {};
    Object.keys(jsonEncodingMap).forEach((fieldKey) => {
        const fieldType = jsonEncodingMap[fieldKey];
        const capitalizedFieldKey = capitalizeFirstLetter(fieldKey);
        jsonObj[capitalizedFieldKey] = encodeValue(fieldType, mappedObj[capitalizedFieldKey]);
    });
    return jsonObj;
}
exports.toJSON = toJSON;
function encodeMap(valueDescription, mapValue) {
    const mapObject = {};
    mapValue.forEach((value, key) => {
        // Use .toString() for keys (key type should have .toString() method)
        const capitalizedKey = capitalizeFirstLetter(key.toString());
        mapObject[capitalizedKey] = encodeValue(valueDescription, value);
    });
    return mapObject;
}
exports.encodeMap = encodeMap;
function decodeMap(keyDescription, valueDescription, jsonMapValue) {
    const mapValue = new Map();
    Object.keys(jsonMapValue).forEach((mapKey) => {
        const mapFieldKey = (keyDescription.type === 'string') ? lowercaseFirstLetter(mapKey) : decodeValue(keyDescription, mapKey);
        mapValue.set(mapFieldKey, decodeValue(valueDescription, jsonMapValue[mapKey]));
    });
    return mapValue;
}
exports.decodeMap = decodeMap;
function encodeObject(objectDescription, objectValue) {
    const resultObject = {};
    Object.keys(objectDescription).forEach((fieldKey) => {
        const valueDescription = objectDescription[fieldKey];
        const capitalizedKey = capitalizeFirstLetter(fieldKey);
        resultObject[capitalizedKey] = encodeValue(valueDescription, objectValue[fieldKey]);
    });
    return resultObject;
}
function encodeArray(valueDescription, arrayValue) {
    return Array.isArray(arrayValue)
        ? arrayValue.map((value) => encodeValue(valueDescription, value))
        : null;
}
function encodeValue(fieldType, fieldValue) {
    switch (fieldType.type) {
        case 'map': {
            // Create a custom object if field is of a map type
            return encodeMap(fieldType.value, fieldValue);
        }
        case 'object': {
            // Create a custom object if field is of an object type
            return encodeObject(fieldType.value, fieldValue);
        }
        case 'array': {
            // Create a custom array if field is of an array type
            return encodeArray(fieldType.value, fieldValue);
        }
        case 'bigint': {
            return fieldValue === undefined ? null : fieldValue;
        }
        case 'buffer': {
            // Marshall buffer as a base64 string
            return (fieldValue === null) ? null : fieldValue.toString('base64');
        }
        case 'address': {
            // Marshall address strings in lowercase
            return fieldValue.toLowerCase();
        }
        default:
            return fieldValue;
    }
}
function capitalizeFirstLetter(str) {
    return str.charAt(0).toUpperCase() + str.slice(1);
}
function lowercaseFirstLetter(str) {
    return str.charAt(0).toLowerCase() + str.slice(1);
}
//# sourceMappingURL=json.js.map